@@ -3,35 +3,38 @@
  *  Module    : main.c
  *  Author    : I. Lea & R. Skrenta
  *  Created   : 1991-04-01
- *  Updated   : 2018-02-15
+ *  Updated   : 2022-11-01
  *  Notes     :
  *
- * Copyright (c) 1991-2019 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
+ * Copyright (c) 1991-2023 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -65,7 +68,7 @@ static t_bool start_any_unread = FALSE;	/* only start if unread news */
 static void create_mail_save_dirs(void);
 static void read_cmd_line_options(int argc, char *argv[]);
 static void show_intro_page(void);
-static void update_index_files(void);
+_Noreturn static void update_index_files(void);
 static void usage(char *theProgname);
 
 
@@ -177,15 +180,21 @@ main(
 	no_write = TRUE;		/* don't allow any writing back during startup */
 
 	if (!batch_mode) {
-#ifdef M_UNIX
-#	ifndef USE_CURSES
+#ifndef USE_CURSES
 		if (!get_termcaps()) {
 			error_message(2, _(txt_screen_init_failed), tin_progname);
 			free_all_arrays();
 			giveup();
 		}
-#	endif /* !USE_CURSES */
-#endif /* M_UNIX */
+#endif /* !USE_CURSES */
+
+		/* exit early at -o without postponed articles */
+		if (post_postponed_and_exit && !count_postponed_articles()) {
+			no_write = TRUE;
+			/* TODO: looks ugly */
+			error_message(0, txt_info_nopostponed);
+			tin_done(EXIT_SUCCESS, NULL);
+		}
 
 		/*
 		 * Init curses emulation
@@ -204,18 +213,35 @@ main(
 		setup_screen();
 	}
 
-	if (!batch_mode || verbose)
+	if (!batch_mode || verbose) {
+		if (!batch_mode && (cLINES < MIN_LINES_ON_TERMINAL || cCOLS < MIN_COLUMNS_ON_TERMINAL)) {
+			ring_bell();
+			tin_done(EXIT_FAILURE, _(txt_screen_too_small_exiting), tin_progname);
+		}
 		wait_message(0, "%s\n", cvers);
+	}
 
 	/*
 	 * Connect to nntp server?
 	 */
 	if (!nntp_server || !*nntp_server)
 		nntp_server = getserverbyfile(NNTP_SERVER_FILE);
-	if (read_news_via_nntp && !read_saved_news && nntp_open()) {
-		nntp_close();
-		free_all_arrays();
-		giveup();
+
+	open_msglog(); /* depends on nntp_server */
+
+	if (read_news_via_nntp && !read_saved_news) {
+		if (use_nntps && tintls_init()) {
+			tintls_exit();
+			free_all_arrays();
+			giveup();
+		}
+
+		if (nntp_open()) {
+			nntp_close(FALSE);
+			tintls_exit();
+			free_all_arrays();
+			giveup();
+		}
 	}
 
 	read_server_config();
@@ -288,7 +314,8 @@ main(
 	read_attributes_file(FALSE);
 	start_groupnum = read_news_active_file();
 #ifdef DEBUG
-	debug_print_active();
+	if (debug & DEBUG_MISC)
+		debug_print_active();
 #endif /* DEBUG */
 
 	/*
@@ -299,7 +326,8 @@ main(
 
 	no_write = TRUE;
 #ifdef DEBUG
-	debug_print_filters();
+	if (debug & DEBUG_FILTER)
+		debug_print_filters();
 #endif /* DEBUG */
 
 	/*
@@ -438,11 +466,15 @@ main(
 
 /*
  * process command line options
- * [01235789beEFijJkKLOtTyY] are unused
+ * [01235789beEFijJKLOtyY] are unused
  * [W] is reserved
- * [BCPU] have been in use at some time, but now are unused
+ * [BCPU] have been in use at some time, but now are unused:
+ *   B BBS mode (M_AMIGA only)
+ *   C count articles
+ *   P purge group index files of articles that no longer exist
+ *   U update index files in background
  */
-#define OPTIONS "46aAcdD:f:g:G:hHI:lm:M:nNop:qQrRs:SuvVwxXzZ"
+#define OPTIONS "46aAcdD:f:g:G:hHI:klm:M:nNop:qQrRs:STuvVwxXzZ"
 
 static void
 read_cmd_line_options(
@@ -570,7 +602,20 @@ read_cmd_line_options(
 				/* FALLTHROUGH */
 
 			case 'I':
-				my_strncpy(index_newsdir, optarg, sizeof(index_newsdir) - 1);
+				joinpath(index_newsdir, sizeof(index_newsdir), optarg, INDEX_NEWSDIR);
+				break;
+
+			case 'k':
+#ifdef NNTPS_ABLE
+				insecure_nntps = TRUE;
+				use_nntps = TRUE;
+#else
+				error_message(2, _(txt_option_not_enabled), "--with-nntps");
+				free_all_arrays();
+				giveup();
+				/* keep lint quiet: */
+				/* NOTREACHED */
+#endif /* NNTPS_ABLE */
 				break;
 
 			case 'l':
@@ -669,6 +714,18 @@ read_cmd_line_options(
 				batch_mode = TRUE;
 				break;
 
+			case 'T':
+#ifdef NNTPS_ABLE
+				use_nntps = TRUE;
+#else
+				error_message(2, _(txt_option_not_enabled), "--with-nntps");
+				free_all_arrays();
+				giveup();
+				/* keep lint quiet: */
+				/* NOTREACHED */
+#endif /* NNTPS_ABLE */
+				break;
+
 			case 'u':	/* update index files */
 				batch_mode = TRUE;
 				update_index = TRUE;
@@ -728,6 +785,18 @@ read_cmd_line_options(
 				exit(EXIT_SUCCESS);
 		}
 	}
+
+#ifdef NNTP_ABLE
+	if (nntp_tcp_port == 0) {
+#ifdef NNTPS_ABLE
+		if (use_nntps)
+			nntp_tcp_port = nntps_tcp_default_port;
+		else
+#endif /* NNTPS_ABLE */
+			nntp_tcp_port = nntp_tcp_default_port;
+	}
+#endif /* NNTP_ABLE*/
+
 	cmdargs = argv;
 	num_cmdargs = optind;
 	max_cmdargs = argc;
@@ -742,9 +811,9 @@ read_cmd_line_options(
 			get_newsrcname(newsrc, sizeof(newsrc), uts.nodename);
 #else
 			char nodenamebuf[256]; /* SUSv2 limit; better use HOST_NAME_MAX */
-#ifdef HAVE_GETHOSTNAME
+#	ifdef HAVE_GETHOSTNAME
 			(void) gethostname(nodenamebuf, sizeof(nodenamebuf));
-#endif /* HAVE_GETHOSTNAME */
+#	endif /* HAVE_GETHOSTNAME */
 			get_newsrcname(newsrc, sizeof(newsrc), nodenamebuf);
 #endif /* HAVE_SYS_UTSNAME_H && HAVE_UNAME */
 		}
@@ -770,14 +839,32 @@ read_cmd_line_options(
 			verbose = FALSE;
 		}
 	}
+	/*
+	 * TODO: also disallow
+	 *       -NM
+	 *       -oN, -oM (at this stage we no longer know if -N or -M was given)
+	 *       -wN, -wM (at this stage we no longer know if -N or -M was given)
+	 */
 	if (post_postponed_and_exit && force_no_post) {
 		wait_message(2, _(txt_useless_combination), "-o", "-x", "-x");
 		force_no_post = FALSE;
 	}
+	if (post_postponed_and_exit && start_any_unread) {
+		wait_message(2, _(txt_useless_combination), "-o", "-z", "-z");
+		start_any_unread = FALSE;
+	}
+	if (post_postponed_and_exit && post_article_and_exit) {
+		wait_message(2, _(txt_useless_combination), "-o", "-w", "-w");
+		post_article_and_exit = FALSE;
+	}
 	if (post_article_and_exit && force_no_post) {
 		wait_message(2, _(txt_useless_combination), "-w", "-x", "-x");
 		force_no_post = FALSE;
 	}
+	if (post_article_and_exit && start_any_unread) {
+		wait_message(2, _(txt_useless_combination), "-w", "-z", "-z");
+		start_any_unread = FALSE;
+	}
 	if (catchup && start_any_unread) {
 		wait_message(2, _(txt_useless_combination), "-c", "-z", "-c");
 		catchup = FALSE;
@@ -864,15 +951,15 @@ usage(
 	error_message(2, _(txt_usage_tin), theProgname);
 
 #if defined(NNTP_ABLE) && defined(INET6)
-		error_message(2, _(txt_usage_force_ipv4));
-		error_message(2, _(txt_usage_force_ipv6));
+	error_message(2, _(txt_usage_force_ipv4));
+	error_message(2, _(txt_usage_force_ipv6));
 #endif /* NNTP_ABLE && INET6 */
 
 #ifdef HAVE_COLOR
-		error_message(2, _(txt_usage_toggle_color));
+	error_message(2, _(txt_usage_toggle_color));
 #endif /* HAVE_COLOR */
 #ifdef NNTP_ABLE
-		error_message(2, _(txt_usage_force_authentication));
+	error_message(2, _(txt_usage_force_authentication));
 #endif /* NNTP_ABLE */
 
 	error_message(2, _(txt_usage_catchup));
@@ -896,6 +983,13 @@ usage(
 	error_message(2, _(txt_usage_help_message));
 	error_message(2, _(txt_usage_help_information), theProgname);
 	error_message(2, _(txt_usage_index_newsdir), index_newsdir);
+
+#ifdef NNTP_ABLE
+#	ifdef NNTPS_ABLE
+	error_message(2, _(txt_usage_use_insecure_nntps));
+#	endif /* NNTPS_ABLE */
+#endif /* NNTP_ABLE */
+
 	error_message(2, _(txt_usage_read_only_active));
 	error_message(2, _(txt_usage_maildir), tinrc.maildir);
 	error_message(2, _(txt_usage_mail_new_news_to_user));
@@ -904,7 +998,11 @@ usage(
 	error_message(2, _(txt_usage_post_postponed_arts));
 
 #ifdef NNTP_ABLE
-	error_message(2, _(txt_usage_port), nntp_tcp_port);
+#	ifdef NNTPS_ABLE
+	error_message(2, _(txt_usage_port), use_nntps ? nntps_tcp_default_port : nntp_tcp_default_port);
+#	else
+	error_message(2, _(txt_usage_port), nntp_tcp_default_port);
+#	endif /* NNTPS_ABLE */
 #endif /* NNTP_ABLE */
 
 	error_message(2, _(txt_usage_dont_check_new_newsgroups));
@@ -918,6 +1016,11 @@ usage(
 	error_message(2, _(txt_usage_read_saved_news));
 	error_message(2, _(txt_usage_savedir), tinrc.savedir);
 	error_message(2, _(txt_usage_save_new_news));
+#ifdef NNTP_ABLE
+#	ifdef NNTPS_ABLE
+	error_message(2, _(txt_usage_use_nntps));
+#	endif /* NNTPS_ABLE */
+#endif /* NNTP_ABLE */
 	error_message(2, _(txt_usage_update_index_files));
 	error_message(2, _(txt_usage_verbose));
 	error_message(2, _(txt_usage_version));
@@ -934,13 +1037,15 @@ usage(
 /*
  * update index files
  */
-static void
+_Noreturn static void
 update_index_files(
 	void)
 {
 	cCOLS = 132;							/* set because curses has not started */
 	create_index_lock_file(lock_file);
 	tinrc.thread_articles = THREAD_NONE;	/* stop threading to run faster */
+	tinrc.sort_article_type = SORT_ARTICLES_BY_NOTHING;
+	tinrc.sort_threads_type = SORT_THREADS_BY_NOTHING;
 	do_update(catchup);
 	tin_done(EXIT_SUCCESS, NULL);
 }
@@ -1040,7 +1145,7 @@ create_mail_save_dirs(
  * already malloc()ed would be a PITA.
  */
 /* coverity[+kill] */
-void
+_Noreturn void
 giveup(
 	void)
 {
@@ -1055,5 +1160,7 @@ giveup(
 		EndWin();
 		Raw(FALSE);
 	}
+	close_msglog();
+
 	exit(EXIT_FAILURE);
 }
