@@ -3,35 +3,38 @@
  *  Module    : string.c
  *  Author    : Urs Janssen <urs@tin.org>
  *  Created   : 1997-01-20
- *  Updated   : 2016-02-25
+ *  Updated   : 2022-11-28
  *  Notes     :
  *
- * Copyright (c) 1997-2016 Urs Janssen <urs@tin.org>
+ * Copyright (c) 1997-2023 Urs Janssen <urs@tin.org>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions and the following disclaimer.
+ *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote
- *    products derived from this software without specific prior written
- *    permission.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
- * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
- * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
- * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * 3. Neither the name of the copyright holder nor the names of its
+ *    contributors may be used to endorse or promote products derived from
+ *    this software without specific prior written permission.
+ *
+ * * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 
 
@@ -69,9 +72,8 @@
  * special ltoa()
  * converts value into a string with a maxlen of digits (usually should be
  * >=4), last char may be one of the following:
- * 'k'ilo, 'M'ega, 'G'iga, 'T'erra, 'P'eta, 'E'xa, 'Z'etta, 'Y'otta,
- * 'X'ona, 'W'eka, 'V'unda, 'U'da (these last 4 are no official SI-prefixes)
- * or 'e' if an error occurs
+ * 'k'ilo, 'M'ega, 'G'iga, 'T'era, 'P'eta, 'E'xa, 'Z'etta, 'Y'otta,
+ * 'R'onna, 'Q'uetta or 'e' if an error occurs
  */
 char *
 tin_ltoa(
@@ -79,7 +81,7 @@ tin_ltoa(
 	int digits)
 {
 	static char buffer[64];
-	static const char power[] = { ' ', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y', 'X', 'W', 'V', 'U', '\0' };
+	static const char power[] = { ' ', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y', 'R', 'Q', '\0' };
 	int len;
 	size_t i = 0;
 
@@ -91,9 +93,15 @@ tin_ltoa(
 	snprintf(buffer, sizeof(buffer), "%"T_ARTNUM_PFMT, value);
 	len = (int) strlen(buffer);
 
-	while (len > digits) {
-		len -= 3;
-		i++;
+	/*
+	 * only shorten if necessary,
+	 * then ensure that the metric prefix fits into the buffer
+	 */
+	if (len > digits) {
+		while (len >= digits) {
+			len -= 3;
+			i++;
+		}
 	}
 
 	if (i >= strlen(power)) {	/* buffer is to small */
@@ -213,11 +221,11 @@ strcasestr(
 	h = haystack;
 	n = needle;
 	while (*haystack) {
-		if (tolower((unsigned char) *h) == tolower((unsigned char) *n)) {
+		if (my_tolower((unsigned char) *h) == my_tolower((unsigned char) *n)) {
 			h++;
 			n++;
 			if (!*n)
-				return haystack;
+				return (char *) haystack;
 		} else {
 			h = ++haystack;
 			n = needle;
@@ -239,11 +247,45 @@ mystrcat(
 		*((*t)++) = *s++;
 		len++;
 	}
-	**t = 0;
+	**t = '\0';
 	return len;
 }
 
 
+/*
+ * get around broken tolower()/toupper() macros on
+ * ancient BSDs (e.g. 4.2, 4.3, 4.3-Tahoe, 4.3-Reno and Net/2)
+ */
+int
+my_tolower(
+	int c)
+{
+#ifdef TOLOWER_BROKEN
+	if (c >= 'A' && c <= 'Z')
+		return (c - 'A' + 'a');
+	else
+		return (c);
+#else
+	return tolower(c);
+#endif /* TOLOWER_BROKEN */
+}
+
+
+int
+my_toupper(
+	int c)
+{
+#ifdef TOUPPER_BROKEN
+	if (c >= 'a' && c <= 'z')
+		return (c - 'a' + 'A');
+	else
+		return (c);
+#else
+	return toupper(c);
+#endif /* TOUPPER_BROKEN */
+}
+
+
 void
 str_lwr(
 	char *str)
@@ -251,7 +293,7 @@ str_lwr(
 	char *dst = str;
 
 	while (*str)
-		*dst++ = (char) tolower((unsigned char) *str++);
+		*dst++ = (char) my_tolower((unsigned char) *str++);
 
 	*dst = '\0';
 }
@@ -368,7 +410,7 @@ atol(
 
 #ifndef HAVE_STRTOL
 /* fix me - put me in tin.h */
-#	define DIGIT(x) (isdigit((unsigned char) x) ? ((x) - '0') : (10 + tolower((unsigned char) x) - 'a'))
+#	define DIGIT(x) (isdigit((unsigned char) x) ? ((x) - '0') : (10 + my_tolower((unsigned char) x) - 'a'))
 #	define MBASE 36
 long
 strtol(
@@ -422,7 +464,7 @@ OUT:
 
 #if !defined(HAVE_STRCASECMP) || !defined(HAVE_STRNCASECMP)
 	/* fix me - put me in tin.h */
-#	define FOLD_TO_UPPER(a)	(toupper((unsigned char) (a)))
+#	define FOLD_TO_UPPER(a)	(my_toupper((unsigned char) (a)))
 #endif /* !HAVE_STRCASECMP || !HAVE_STRNCASECMP */
 /*
  * strcmp that ignores case
@@ -624,7 +666,7 @@ expand_tab(
 
 #define SH_FORMAT(c)	if (++result >= (int) len) \
 				break; \
-			*dst++ = c
+			*dst++ = (char) c
 
 #define SH_SINGLE "\\\'"
 #define SH_DOUBLE "\\\'\"`$"
@@ -690,11 +732,11 @@ sh_format(
 				 * the whole string rather than escaping individual chars.
 				 */
 				if (quote == '"') {
-					fix = (strchr(SH_DOUBLE, *src) != 0);
+					fix = (strchr(SH_DOUBLE, *src) != NULL);
 				} else if (quote == '\'') {
-					fix = (strchr(SH_SINGLE, *src) != 0);
+					fix = (strchr(SH_SINGLE, *src) != NULL);
 				} else
-					fix = (strchr(SH_META, *src) != 0);
+					fix = (strchr(SH_META, *src) != NULL);
 				if (fix) {
 					SH_FORMAT('\\');
 				}
@@ -755,7 +797,7 @@ strrstr(
 	const char *ptr;
 	size_t slen, plen;
 
-	if ((str != 0) && (pat != 0)) {
+	if ((str != NULL) && (pat != NULL)) {
 		slen = strlen(str);
 		plen = strlen(pat);
 
@@ -795,7 +837,7 @@ char2wchar_t(
 		return NULL;
 	}
 
-	wstr = my_malloc(sizeof(wchar_t) * (len + 1));
+	wstr = my_calloc(1, sizeof(wchar_t) * (len + 1));
 	mbstowcs(wstr, test, len + 1);
 	free(test);
 
@@ -875,7 +917,7 @@ wcspart(
 
 		gap = columns - wcswidth(wbuf, wcslen(wbuf) + 1);
 		assert(gap >= 0);
-		wbuf = my_realloc(wbuf, sizeof(wchar_t) * (wcslen(wbuf) + gap + 1));
+		wbuf = my_realloc(wbuf, sizeof(wchar_t) * (wcslen(wbuf) + (size_t) gap + 1));
 		ptr = wbuf + wcslen(wbuf); /* set ptr again to end of wbuf */
 
 		while (gap-- > 0)
@@ -997,7 +1039,7 @@ abbr_groupname(
 
 	return new_grpname;
 }
-#endif /* MULTIBYTE_ABLE && !NOLOCALE */
+#endif /* MULTIBYTE_ABLE && !NO_LOCALE */
 
 
 /*
@@ -1025,7 +1067,7 @@ strwidth(
 #define TRUNC_TAIL	"..."
 /*
  * shortens 'mesg' that it occupies at most 'len' screen positions.
- * If it was nessary to truncate 'mesg', " ..." is appended to the
+ * If it was necessary to truncate 'mesg', " ..." is appended to the
  * resulting string (still 'len' screen positions wide).
  * The resulting string is stored in 'buf'.
  */
@@ -1056,7 +1098,7 @@ strunc(
 		tmp = my_strdup(message);
 	else {
 		tmp = my_malloc(len + 1);
-		snprintf(tmp, len + 1, "%-.*s%s", len - 3, message, TRUNC_TAIL);
+		snprintf(tmp, (size_t) (len + 1), "%-.*s%s", len - 3, message, TRUNC_TAIL);
 	}
 
 	return tmp;
@@ -1092,7 +1134,7 @@ wstrunc(
 			tail = char2wchar_t(TRUNC_TAIL);
 
 		len_tail = tail ? wcslen(tail) : 0;
-		wtmp2 = wcspart(wtmp, len - len_tail, FALSE);
+		wtmp2 = wcspart(wtmp, (int) ((size_t) len - len_tail), FALSE);
 		free(wtmp);
 		wtmp = my_realloc(wtmp2, sizeof(wchar_t) * (wcslen(wtmp2) + len_tail + 1));	/* wtmp2 isn't valid anymore and doesn't have to be free()ed */
 		if (!tail)
@@ -1202,35 +1244,77 @@ normalize(
 		int32_t needed, norm_len;
 		UChar *ustr, *norm;
 		UErrorCode status = U_ZERO_ERROR;
+#		ifdef HAVE_UNICODE_UNORM2_H
+		static const char *uni_name[] = {"nfc", "nfkc", "nfkc_cf"}; /* */
+		const char *uni_namep;
+		UNormalization2Mode mode;
+#		else
 		UNormalizationMode mode;
-
-		switch (tinrc.normalization_form) {
-			case NORMALIZE_NFD:
-				mode = UNORM_NFD;
-				break;
-
-			case NORMALIZE_NFC:
-				mode = UNORM_NFC;
-				break;
-
-			case NORMALIZE_NFKD:
-				mode = UNORM_NFKD;
-				break;
-
-			case NORMALIZE_NFKC:
-			default:
-				mode = UNORM_NFKC;
-		}
+#		endif /* !HAVE_UNICODE_UNORM2_H */
 
 		/* convert to UTF-16 which is used internally by ICU */
 		if ((ustr = char2UChar(tmp)) == NULL) /* something went wrong, return the original string (as valid UTF8) */
 			return tmp;
 
+		switch (tinrc.normalization_form) {
+			case NORMALIZE_NFD:
+#		ifdef HAVE_UNICODE_UNORM2_H
+				uni_namep = uni_name[0];
+				mode = UNORM2_DECOMPOSE;
+#		else
+				mode = UNORM_NFD;
+#		endif /* HAVE_UNICODE_UNORM2_H */
+				break;
+
+			case NORMALIZE_NFC:
+#		ifdef HAVE_UNICODE_UNORM2_H
+				uni_namep = uni_name[0];
+				mode = UNORM2_COMPOSE;
+#		else
+				mode = UNORM_NFC;
+#		endif /* HAVE_UNICODE_UNORM2_H */
+				break;
+
+			case NORMALIZE_NFKD:
+#		ifdef HAVE_UNICODE_UNORM2_H
+				uni_namep = uni_name[1];
+				mode = UNORM2_DECOMPOSE;
+#		else
+				mode = UNORM_NFKD;
+#		endif /* HAVE_UNICODE_UNORM2_H */
+				break;
+#		ifdef HAVE_UNICODE_UNORM2_H
+			case NORMALIZE_NFKC_CF:
+				uni_namep = uni_name[2];
+				mode = UNORM2_COMPOSE;
+				break;
+#		endif /* HAVE_UNICODE_UNORM2_H */
+
+			case NORMALIZE_NFKC:
+			default:
+#		ifdef HAVE_UNICODE_UNORM2_H
+				uni_namep = uni_name[1];
+				mode = UNORM2_COMPOSE;
+#		else
+				mode = UNORM_NFKC;
+#		endif /* HAVE_UNICODE_UNORM2_H */
+		}
+
+#		ifdef HAVE_UNICODE_UNORM2_H
+		needed = unorm2_normalize(unorm2_getInstance(NULL, uni_namep, mode, &status), ustr, -1, NULL, 0, &status);
+#		else
 		needed = unorm_normalize(ustr, -1, mode, 0, NULL, 0, &status);
+#		endif /* HAVE_UNICODE_UNORM2_H */
+
 		status = U_ZERO_ERROR;		/* reset status */
 		norm_len = needed + 1;
 		norm = my_malloc(sizeof(UChar) * norm_len);
+#		ifdef HAVE_UNICODE_UNORM2_H
+		(void) unorm2_normalize(unorm2_getInstance(NULL, uni_namep, mode, &status), ustr, -1, norm, norm_len, &status);
+#		else
 		(void) unorm_normalize(ustr, -1, mode, 0, norm, norm_len, &status);
+#		endif /* HAVE_UNICODE_UNORM2_H */
+
 		if (U_FAILURE(status)) {
 			/* something went wrong, return the original string (as valid UTF8) */
 			free(ustr);
@@ -1273,6 +1357,7 @@ normalize(
 				mode = UNINORM_NFKC;
 		}
 		buf = (char *) u8_normalize(mode, (uint8_t *) tmp, strlen(tmp) + 1, NULL, &olen);
+		free(tmp);
 		return buf;
 	}
 #		else
@@ -1301,21 +1386,22 @@ fmt_string(
 	const char *fmt,
 	...)
 {
-	char *str;
 	va_list ap;
+	size_t size = LEN;
+	char *str = my_malloc(size);
+	int used;
 
-	va_start(ap, fmt);
-#ifdef HAVE_VASPRINTF
-	if (vasprintf(&str, fmt, ap) == -1)	/* something went wrong */
-#endif /* HAVE_VASPRINTF */
-	{
-		size_t size = LEN;
+	while (1) {
+		va_start(ap, fmt);
+		used = vsnprintf(str, size, fmt, ap);
+		va_end(ap);
 
-		str = my_malloc(size);
-		/* TODO: realloc str if necessary */
-		vsnprintf(str, size, fmt, ap);
+		if (used > 0 && used < (int) size)
+			break;
+
+		size <<= 1;
+		str = my_realloc(str, size);
 	}
-	va_end(ap);
 
 	return str;
 }
@@ -1426,13 +1512,13 @@ parse_format_string(
 		tmp_date_str[0] = '\0';
 		d_fmt = tmp_date_str;
 		if (*in > '0' && *in <= '9') {
-			len = atoi(in);
+			len = (size_t) atoi(in);
 			for (; *in >= '0' && *in <= '9'; in++)
 				;
 		}
 		if (*in == ',') {
 			if (*++in > '0' && *in <= '9') {
-				len2 = atoi(in);
+				len2 = (size_t) atoi(in);
 				for (; *in >= '0' && *in <= '9'; in++)
 					;
 			}
@@ -1453,7 +1539,7 @@ parse_format_string(
 				endp = startp = tmpp;
 
 			if (endp) {
-				tmplen = endp - in;
+				tmplen = (size_t) (endp - in);
 
 				for (in++; *in && --tmplen; in++)
 					*d_fmt++ = *in;
@@ -1480,9 +1566,8 @@ parse_format_string(
 				if (cCOLS > min_cols && !(flags & GRP_DESC) && signal_context == cSelect) {
 					flags |= GRP_DESC;
 					fmt->show_grpdesc = TRUE;
-					if (len) {
+					if (len)
 						fmt->len_grpdesc = len;
-					}
 				} else
 					out -= 2;
 				break;
@@ -1491,7 +1576,7 @@ parse_format_string(
 				/* Date */
 				if (cCOLS > min_cols && (!(flags & DATE) && (signal_context == cGroup || signal_context == cThread))) {
 					flags |= DATE;
-					if (strlen(tmp_date_str))
+					if (*tmp_date_str)
 						strcpy(fmt->date_str, tmp_date_str);
 					else
 						STRCPY(fmt->date_str, curr_group->attribute->date_format);
@@ -1501,7 +1586,7 @@ parse_format_string(
 #if defined(MULTIBYTE_ABLE) && !defined(NO_LOCALE)
 						if ((wtmp = char2wchar_t(buf)) != NULL) {
 							if (wcstombs(tmp, wtmp, sizeof(tmp) - 1) != (size_t) -1) {
-								fmt->len_date = strwidth(tmp);
+								fmt->len_date = (size_t) strwidth(tmp);
 							}
 							free(wtmp);
 						}
@@ -1617,9 +1702,8 @@ parse_format_string(
 				/* Subject */
 				if (cCOLS > min_cols && !(flags & SUBJECT) && signal_context == cGroup) {
 					flags |= SUBJECT;
-					if (len) {
+					if (len)
 						fmt->len_subj = len;
-					}
 				} else
 					out -= 2;
 				break;
@@ -1639,9 +1723,8 @@ parse_format_string(
 				if (cCOLS > min_cols && !(flags & THREAD_TREE) && signal_context == cThread) {
 					flags |= THREAD_TREE;
 					show_subject = TRUE;
-					if (len) {
+					if (len)
 						fmt->len_subj = len;
-					}
 				} else
 					out -= 2;
 				break;
@@ -1693,7 +1776,7 @@ parse_format_string(
 				fmt->flags_offset = tinrc.draw_arrow ? 2 : 0;
 				fmt->ucnt_offset = tinrc.draw_arrow ? 10 : 8;
 				fmt->len_grpname_dsc = 32;
-				fmt->len_grpname_max = cCOLS - cnt - 1;
+				fmt->len_grpname_max = (size_t) cCOLS - cnt - 1;
 				fmt->len_ucnt = 5;
 				break;
 
@@ -1719,36 +1802,36 @@ parse_format_string(
 		}
 
 		if (flags & SUBJECT)
-			fmt->len_from = (cCOLS - cnt - 1) / 3;
+			fmt->len_from = ((size_t) cCOLS - cnt - 1) / 3;
 		else
-			fmt->len_from = (cCOLS - cnt - 1);
+			fmt->len_from = ((size_t) cCOLS - cnt - 1);
 
-		fmt->len_subj = cCOLS - fmt->len_from - cnt - 1;
+		fmt->len_subj = (size_t) cCOLS - fmt->len_from - cnt - 1;
 	} else {
 		if (flags & (GRP_NAME | GRP_DESC))
-			fmt->len_grpname_max = cCOLS - cnt - 1;
+			fmt->len_grpname_max = (size_t) cCOLS - cnt - 1;
 
 		if (!show_description && !(flags & GRP_NAME))
 			fmt->len_grpname_max = 0;
 
-		if (flags & DATE && fmt->len_date > (cCOLS - cnt - 1))
-			fmt->len_date = (cCOLS - cnt - 1);
+		if (flags & DATE && fmt->len_date > ((size_t) cCOLS - cnt - 1))
+			fmt->len_date = ((size_t) cCOLS - cnt - 1);
 
-		if (flags & DATE && (!fmt->len_date_max || fmt->len_date_max > (cCOLS - cnt - 1)))
+		if (flags & DATE && (!fmt->len_date_max || fmt->len_date_max > ((size_t) cCOLS - cnt - 1)))
 			fmt->len_date_max = fmt->len_date;
 
-		if (flags & FROM && (!fmt->len_from || fmt->len_from > (cCOLS - fmt->len_date_max - cnt - 1))) {
+		if (flags & FROM && (!fmt->len_from || fmt->len_from > ((size_t) cCOLS - fmt->len_date_max - cnt - 1))) {
 			if (flags & (SUBJECT | THREAD_TREE)) {
 				if (fmt->len_subj)
-					fmt->len_from = cCOLS - fmt->len_date_max - fmt->len_subj - cnt - 1;
+					fmt->len_from = (size_t) cCOLS - fmt->len_date_max - fmt->len_subj - cnt - 1;
 				else
-					fmt->len_from = (cCOLS - fmt->len_date_max - cnt - 1) / 3;
+					fmt->len_from = ((size_t) cCOLS - fmt->len_date_max - cnt - 1) / 3;
 			} else
-				fmt->len_from = (cCOLS - fmt->len_date_max - cnt - 1);
+				fmt->len_from = ((size_t) cCOLS - fmt->len_date_max - cnt - 1);
 		}
 
-		if (flags & (SUBJECT | THREAD_TREE) && (!fmt->len_subj || fmt->len_subj > (cCOLS - fmt->len_from - fmt->len_date_max - cnt - 1)))
-			fmt->len_subj = (cCOLS - fmt->len_from - fmt->len_date_max - cnt - 1);
+		if (flags & (SUBJECT | THREAD_TREE) && (!fmt->len_subj || fmt->len_subj > ((size_t) cCOLS - fmt->len_from - fmt->len_date_max - cnt - 1)))
+			fmt->len_subj = ((size_t) cCOLS - fmt->len_from - fmt->len_date_max - cnt - 1);
 	}
 }
 
